(* HEXCONV.PAS                             *)
(* HEX file converter for dsPIC bootloader *)
(* (c)2006,2007 Cepstrum                   *)
(* http://www.cepstrum.co.jp/              *)

{$G1} (* redirect input *)
{$P1} (* redirect output *)
{$U+} (* user interrupt enable *)
{$R+} (* array range check enable *)

const
  copyright='(c)2006,2007 Cepstrum';
  url      ='http://www.cepstrum.co.jp/';

type
  STR4=string[4];

var
  dat   : array[0..8191] of string[4];
  dummy : string[255];

(*----------------------------------------------*)

function int2hex(x : integer) : STR4;
  function b2c(b : byte) : char;
    begin
      b2c:=copy('0123456789ABCDEF', b+1, 1);
    end;
  var
    s : STR4;
  begin
    s[4]:=b2c(x mod 16);
    x:=x div 16;
    s[3]:=b2c(x mod 16);
    x:=x div 16;
    s[2]:=b2c(x mod 16);
    x:=x div 16;
    s[1]:=b2c(x);
    int2hex:=s;
  end;

(*----------------------------------------------*)

function str2byte(c1 : char; c2 : char) : byte;
  var
    i, j : byte;
  begin
    i:=ord(c1);
    if 65<=i then       (* if 'A'<=c1 *)
      i:=i-55
    else
      i:=i-48;
    j:=ord(c2);
    if 65<=j then       (* if 'A'<=c2 *)
      j:=j-55
    else
      j:=j-48;
    str2byte:=16*i+j;
  end;

(*----------------------------------------------*)

procedure debugout;
  var
    i, j : integer;
  begin
    for i:=0 to 1023 do
      begin
        j:=8*i;
        write(int2hex(j), ' : ');
        write(copy(dat[j+1], 1, 2));
        write(copy(dat[j+0], 3, 2));
        write(copy(dat[j+0], 1, 2), ' ');
        write(copy(dat[j+3], 1, 2));
        write(copy(dat[j+2], 3, 2));
        write(copy(dat[j+2], 1, 2), ' ');
        write(copy(dat[j+5], 1, 2));
        write(copy(dat[j+4], 3, 2));
        write(copy(dat[j+4], 1, 2), ' ');
        write(copy(dat[j+7], 1, 2));
        write(copy(dat[j+6], 3, 2));
        write(copy(dat[j+6], 1, 2), ' ');
        writeln;
      end;
  end;

(*----------------------------------------------*)

var
  adrh    : integer;
  adrl    : integer;
  sumh    : array[0..127] of byte;
  summ    : array[0..127] of byte;
  suml    : array[0..127] of byte;
  strbuf  : string[255];
  c1, c2  : char;
  b       : byte;
  bytlen  : integer;
  wrdlen  : integer;
  rectyp  : integer;
  tail    : integer;
  h_opt   : boolean;
  d_opt   : boolean;
  a_opt   : boolean;
  i, j, k : integer;

begin

  dummy:=copyright+' '+url;

  h_opt:=false;
  d_opt:=false;
  a_opt:=false;

  for i:=1 to paramcount do
    begin
      if (pos(paramstr(i), '/h-h/H-H/?-?') mod 2)=1 then h_opt:=true;
      if (pos(paramstr(i), '/d-d/D-D') mod 2)=1 then d_opt:=true;
      if (pos(paramstr(i), '/a-a/A-A') mod 2)=1 then a_opt:=true;
    end;

  if h_opt then
    begin
      writeln('usage   : hexconv [-d] < infile.hex > outfile.txt');
      writeln('option  :  -d = debug output mode');
      writeln('           -a = add dummy character for delay');
      write  ('(c)2006,2007 Cepstrum');
      halt;
    end;

  for i:=0 to 8191 do
    dat[i]:='FFFF';

  for i:=0 to 127 do
    begin
      sumh[i]:=0;
      summ[i]:=0;
      suml[i]:=0;
    end;

  while not eof do
    begin
      readln(strbuf);
      for i:=1 to length(strbuf) do
        strbuf[i]:=upcase(strbuf[i]);
      if (length(strbuf)<>0) and (strbuf[1]<>';') then
        begin
          bytlen:=str2byte(strbuf[2], strbuf[3]);
          wrdlen:=bytlen div 2;
          val(copy(strbuf, 8, 2), rectyp, i);
          if rectyp=4 then
            begin
              adrh:=str2byte(strbuf[10], strbuf[11]);
              adrh:=128*adrh+(str2byte(strbuf[12], strbuf[13]) div 2);
            end
          else if adrh=0 then
            begin
              adrl:=str2byte(strbuf[4], strbuf[5]);
              adrl:=128*adrl+(str2byte(strbuf[6], strbuf[7]) div 2);
              for i:=1 to wrdlen do
                begin
                  dat[adrl+i-1]:=copy(strbuf, 6+4*i, 4);
                end;
            end;
        end;
    end;

  for i:=0 to 127 do
    for j:=0 to 31 do
      begin
        k:=64*i+2*j;
        b:=str2byte(copy(dat[k], 1, 1), copy(dat[k], 2, 1));
        suml[i]:=suml[i] xor b;
        b:=str2byte(copy(dat[k], 3, 1), copy(dat[k], 4, 1));
        summ[i]:=summ[i] xor b;
        b:=str2byte(copy(dat[k+1], 1, 1), copy(dat[k+1], 2, 1));
        sumh[i]:=sumh[i] xor b;
      end;

  if d_opt then
    begin
      debugout;
      halt;
    end;

  tail:=8192;
  i:=8191;
  while dat[i]='FFFF' do
    begin
      tail:=i;
      i:=i-1;
    end;
  tail:=64*(((i-1) div 64)+1);

  writeln('U');
  i:=0;
  j:=0;
  while i<tail do
    begin
      if (i mod 64)=0 then
        write('$');
      write(copy(dat[i+1], 1, 2));
      write(copy(dat[i  ], 3, 2));
      write(copy(dat[i  ], 1, 2));
      i:=i+2;
      if (i mod 64)=0 then
        begin
          write(copy(int2hex(sumh[j]), 3, 2));
          write(copy(int2hex(summ[j]), 3, 2));
          write(copy(int2hex(suml[j]), 3, 2));
          if a_opt then
            for k:=1 to 10 do write(#$FF);
          writeln;
          j:=j+1;
        end;
    end;
  writeln('Z');

  write(trm, 'last address = 0x', int2hex(tail-1));

end.




